# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Mathematical and geometrical functions used to compute morphometrics."""
import logging
import math
from itertools import combinations

import numpy as np
from scipy.spatial import ConvexHull
from scipy.spatial.distance import cdist

try:
    # The QhulError was moved in scipy >= 1.8 so if the import fails the old location is imported
    from scipy.spatial import QhullError
except ImportError:  # pragma: no cover
    from scipy.spatial.qhull import QhullError

from neurom.core.dataformat import COLS

L = logging.getLogger(__name__)


def vector(p1, p2):
    """Compute vector between two 3D points.

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        3-vector from p1 - p2
    """
    return np.subtract(p1[COLS.XYZ], p2[COLS.XYZ])


def linear_interpolate(p1, p2, fraction):
    """Returns the point p satisfying: p1 + fraction * (p2 - p1)."""
    return np.array(
        (
            p1[0] + fraction * (p2[0] - p1[0]),
            p1[1] + fraction * (p2[1] - p1[1]),
            p1[2] + fraction * (p2[2] - p1[2]),
        )
    )


def interpolate_radius(r1, r2, fraction):
    """Interpolate the radius between two values.

    Calculate the radius that corresponds to a point P that lies at a fraction of the length
    of a cut cone P1P2 where P1, P2 are the centers of the circles that bound the shape with radii
    r1 and r2 respectively.

    Args:
        r1: float
            Radius of the first node of the segment.
        r2: float
            Radius of the second node of the segment
        fraction: float
            The fraction at which the interpolated radius is calculated.

    Returns: float
        The interpolated radius.

    Note: The fraction is assumed from point P1, not from point P2.
    """

    def f(a, b, c):
        """Returns the length of the interpolated radius calculated using similar triangles."""
        return a + c * (b - a)

    return f(r2, r1, 1.0 - fraction) if r1 > r2 else f(r1, r2, fraction)


def interval_lengths(points, prepend_zero=False):
    """Returns the list of distances between consecutive points.

    Args:
        points: a list of np.array of 3D points
        prepend_zero (bool): if True, the returned array will start with a zero
    """
    intervals = np.linalg.norm(np.diff(np.asarray(points)[:, COLS.XYZ], axis=0), axis=1)
    if prepend_zero:
        return np.insert(intervals, 0, 0)
    return intervals


def path_fraction_id_offset(points, fraction, relative_offset=False):
    """Find segment by fractional offset.

    Find the segment which corresponds to the fraction
    of the path length along the piecewise linear curve which
    is constructed from the set of points.

    Args:
        points: an iterable of indexable objects with indices
        0, 1, 2 correspoding to 3D cartesian coordinates
        fraction: path length fraction (0.0 <= fraction <= 1.0)
        relative_offset: return absolute or relative segment distance

    Returns:
        (segment ID, segment offset) pair.
    """
    if not 0.0 <= fraction <= 1.0:
        raise ValueError("Invalid fraction: %.3f" % fraction)
    lengths = interval_lengths(points)
    cum_lengths = np.cumsum(lengths)
    offset = cum_lengths[-1] * fraction
    seg_id = np.argmin(cum_lengths < offset)
    if seg_id > 0:
        offset -= cum_lengths[seg_id - 1]
    if relative_offset:
        offset /= lengths[seg_id]
    return seg_id, offset


def path_fraction_point(points, fraction):
    """Find coordinates by fractional offset.

    Computes the point which corresponds to the fraction
    of the path length along the piecewise linear curve which
    is constructed from the set of points.

    Args:
        points: an iterable of indexable objects with indices
        0, 1, 2 correspoding to 3D cartesian coordinates
        fraction: path length fraction (0 <= fraction <= 1)

    Returns:
        The 3D coordinates of the aforementioned point
    """
    seg_id, offset = path_fraction_id_offset(points, fraction, relative_offset=True)
    return linear_interpolate(points[seg_id], points[seg_id + 1], offset)


def scalar_projection(v1, v2):
    """Compute the scalar projection of v1 upon v2.

    Args:
        v1, v2: iterable
        indices 0, 1, 2 corresponding to cartesian coordinates

    Returns:
        3-vector of the projection of point p onto the direction of v
    """
    return np.dot(v1, v2) / np.linalg.norm(v2)


def vector_projection(v1, v2):
    """Compute the vector projection of v1 upon v2.

    Args:
        v1, v2: iterable
        indices 0, 1, 2 corresponding to cartesian coordinates

    Returns:
        3-vector of the projection of point p onto the direction of v
    """
    return scalar_projection(v1, v2) * v2 / np.linalg.norm(v2)


def dist_point_line(p, l1, l2):
    """Compute the orthogonal distance between a line and a point.

    The line is that which passes through the points l1 and l2.

    Args:
        p: iterable
            indices 0, 1, 2 correspond to cartesian coordinates
        l1: iterable
            indices 0, 1, 2 correspond to cartesian coordinates
        l2: iterable
            indices 0, 1, 2 correspond to cartesian coordinates
    """
    cross_prod = np.cross(l2 - l1, p - l1)
    return np.linalg.norm(cross_prod) / np.linalg.norm(l2 - l1)


def point_dist2(p1, p2):
    """Compute the square of the euclidian distance between two 3D points.

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        The square of the euclidian distance between the points.
    """
    v = vector(p1, p2)
    return np.dot(v, v)


def point_dist(p1, p2):
    """Compute the euclidian distance between two 3D points.

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        The euclidian distance between the points.
    """
    return np.sqrt(point_dist2(p1, p2))


def angle_3points(p0, p1, p2):
    """Compute the angle in radians between three 3D points.

    Calculated as the angle between p1-p0 and p2-p0.

    Args:
        p0, p1, p2:  indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        Angle in radians between (p1-p0) and (p2-p0).
        0.0 if p0==p1 or p0==p2.
    """
    vec1 = vector(p1, p0)
    vec2 = vector(p2, p0)
    return math.atan2(np.linalg.norm(np.cross(vec1, vec2)), np.dot(vec1, vec2))


def angle_between_vectors(p1, p2):
    """Computes the angle in radians between vectors 'p1' and 'p2'.

    Normalizes the input vectors and computes the relative angle
    between them.
    """
    if np.equal(p1, p2).all():
        return 0.0
    v1 = p1 / np.linalg.norm(p1)
    v2 = p2 / np.linalg.norm(p2)
    return np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))


def angle_between_projections(p1, p2):
    """Angle between the projections p1 and p2 (2d vectors)."""
    ang1 = np.arctan2(*p1[::-1])
    ang2 = np.arctan2(*p2[::-1])
    return ang1 - ang2


def elevation_from_vector(vec):
    """Return the elevation of a vector."""
    norm_vector = np.linalg.norm(vec)

    if norm_vector >= np.finfo(type(norm_vector)).eps:
        return np.arcsin(np.clip(vec[COLS.Y] / norm_vector, -1.0, 1.0))
    raise ValueError("Norm of vector between soma center and section is almost zero.")


def azimuth_from_vector(vec):
    """Return the azimuth of a vector."""
    return np.arctan2(vec[COLS.Z], vec[COLS.X])


def spherical_from_vector(vec):
    """Return the spherical coordinates of a vector: elevation and azimuth.

    .. note::

        * the elevation is the angle between the vector and its projection on the XZ plane.
        * the azimuth is the angle between the X axis and the projection of the vector on the XZ
          plane.

    .. warning:: This frame is not the usual spherical frame (see :ref:`spherical_coordinates`).
    """
    # Azimuth is in [-pi, pi]
    azimuth = azimuth_from_vector(vec)

    # Elevation is in [-pi/2, pi/2]
    elevation = elevation_from_vector(vec)

    return np.array([elevation, azimuth])


def vector_from_spherical(elevation, azimuth, radius=1.0):
    """Return a vector from the frame center to the point in given direction and given radius.

    .. warning:: The frame is not the usual spherical frame (see :ref:`spherical_coordinates`).
    """
    x = np.cos(elevation) * np.cos(azimuth)
    y = np.sin(elevation)
    z = np.cos(elevation) * np.sin(azimuth)

    return radius * np.array([x, y, z])


def angles_to_pi_interval(angles, scale=1.0):
    """Convert any angles into the ]-scale * pi, scale * pi] interval."""
    mod_angle = np.fmod(angles, 2.0 * scale * np.pi)
    mod_angle = np.where(mod_angle <= -scale * np.pi, mod_angle + 2 * scale * np.pi, mod_angle)
    mod_angle = np.where(mod_angle > scale * np.pi, mod_angle - 2 * scale * np.pi, mod_angle)
    return mod_angle


def polygon_diameter(points):
    """Compute the maximun euclidian distance between any two points in a list of points."""
    return max(point_dist(p0, p1) for (p0, p1) in combinations(points, 2))


def average_points_dist(p0, p_list):
    """Computes the average distance between a list of points and a given point p0."""
    return np.mean(list(point_dist(p0, p1) for p1 in p_list))


def path_distance(points):
    """Compute the path distance from given set of points."""
    return interval_lengths(points).sum()


def segment_length(seg):
    """Return the length of a segment.

    Returns: Euclidian distance between centres of points in seg
    """
    return point_dist(seg[0], seg[1])


def segment_length2(seg):
    """Return the square of the length of a segment.

    Returns: Square of Euclidian distance between centres of points in seg
    """
    return point_dist2(seg[0], seg[1])


def segment_radius(seg):
    """Return the mean radius of a segment.

    Returns: arithmetic mean of the radii of the points in seg
    """
    return (seg[0][COLS.R] + seg[1][COLS.R]) / 2.0


def segment_x_coordinate(seg):
    """Return the mean x coordinate of a segment.

    Returns: arithmetic mean of the x coordinates of the points in seg
    """
    return (seg[0][COLS.X] + seg[1][COLS.X]) / 2.0


def segment_y_coordinate(seg):
    """Return the mean y coordinate of a segment.

    Returns: arithmetic mean of the y coordinates of the points in seg
    """
    return (seg[0][COLS.Y] + seg[1][COLS.Y]) / 2.0


def segment_z_coordinate(seg):
    """Return the mean z coordinate of a segment.

    Returns: arithmetic mean of the z coordinates of the points in seg
    """
    return (seg[0][COLS.Z] + seg[1][COLS.Z]) / 2.0


def segment_radial_dist(seg, pos):
    """Return the radial distance of a tree segment to a given point.

    The radial distance is the euclidian distance between the mid-point of
    the segment and the point in question.

    Arguments:
        seg: tree segment

        pos: origin to which distances are measured. It must have at lease 3
        components. The first 3 components are (x, y, z).
    """
    return point_dist(pos, np.divide(np.add(seg[0], seg[1]), 2.0))


def segment_area(seg):
    """Compute the surface area of a segment.

    Approximated as a conical frustum. Does not include the surface area
    of the bounding circles.
    """
    r0 = seg[0][COLS.R]
    r1 = seg[1][COLS.R]
    h2 = point_dist2(seg[0], seg[1])
    return math.pi * (r0 + r1) * math.sqrt((r0 - r1) ** 2 + h2)


def segment_volume(seg):
    """Compute the volume of a segment.

    Approximated as a conical frustum.
    """
    r0 = seg[0][COLS.R]
    r1 = seg[1][COLS.R]
    h = point_dist(seg[0], seg[1])
    return math.pi * h * ((r0 * r0) + (r0 * r1) + (r1 * r1)) / 3.0


def taper_rate(p0, p1):
    """Compute the taper rate between points p0 and p1.

    Args:
        p0, p1: iterables with first 4 components containing (x, y, z, r)

    Returns:
        The taper rate, defined as the absolute value of the difference in
        the diameters of p0 and p1 divided by the euclidian distance
        between them.
    """
    return 2 * abs(p0[COLS.R] - p1[COLS.R]) / point_dist(p0, p1)


def segment_taper_rate(seg):
    """Compute the taper rate of a segment.

    Returns:
        The taper rate, defined as the absolute value of the difference in
        the diameters of the segment's two points divided by the euclidian
        distance between them.
    """
    return taper_rate(seg[0], seg[1])


def pca(points):
    """Estimate the principal components of the covariance on the given point cloud.

    Args:
        points: A numpy array of points of the form ((x1,y1,z1), (x2, y2, z2)...)

    Returns:
        Eigenvalues and respective eigenvectors
    """
    return np.linalg.eig(np.cov(points.transpose()))


def sphere_area(r):
    """Compute the area of a sphere with radius r."""
    return 4.0 * math.pi * r**2


# Useful alias for path_distance
section_length = path_distance


def principal_direction_extent(points):
    """Calculate the extent of a set of 3D points.

    The extent is defined as the maximum distance between the projections on the principal
    directions of the covariance matrix of the points.

    Args:
        points : a 2D numpy array of points with 2 or 3 columns for (x, y, z)

    Returns:
        extents : the extents for each of the eigenvectors of the cov matrix

    Note:
        Direction extents are ordered from largest to smallest.
    """
    # pca can be biased by duplicate points
    points = np.unique(points, axis=0)

    # center the points around 0.0
    points -= np.mean(points, axis=0)

    # principal components
    _, eigenvectors = pca(points)

    # for each eigenvector calculate the scalar projection of the points on it (n_points, n_eigv)
    scalar_projections = points.dot(eigenvectors)

    # range of the projections (abs(max - min)) along each column (eigenvector)
    extents = np.ptp(scalar_projections, axis=0)

    descending_order = np.argsort(extents)[::-1]

    return extents[descending_order]


def convex_hull(points):
    """Get the convex hull from an array of points.

    Returns:
        scipy.spatial.ConvexHull object if successful, otherwise None
    """
    if len(points) == 0:
        L.exception("Failure to compute convex hull because there are no points")
        return None

    try:
        return ConvexHull(points)
    except QhullError:
        L.exception("Failure to compute convex hull because of geometrical degeneracy.")
        return None


def aspect_ratio(points):
    """Computes the min/max ratio of the principal direction extents."""
    extents = principal_direction_extent(points)
    return float(extents.min() / extents.max())


def circularity(points):
    """Computes circularity as 4 * pi * area / perimeter^2.

    Note: For 2D points, ConvexHull.volume corresponds to its area and ConvexHull.area
        to its perimeter.
    """
    hull = convex_hull(points)
    return 4.0 * np.pi * hull.volume / hull.area**2


def shape_factor(points):
    """Computes area over max pairwise distance squared.

    Defined in doi: 10.1109/ICoAC44903.2018.8939083

    Note: For 2D points, ConvexHull.volume corresponds to its area.
    """
    hull = convex_hull(points)
    hull_points = points[hull.vertices]

    max_pairwise_distance = np.max(cdist(hull_points, hull_points))

    return hull.volume / max_pairwise_distance**2



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Geometrical Operations for NeuroM."""

import logging

import numpy as np

import neurom.morphmath
from neurom.core.dataformat import COLS
from neurom.geom.transform import rotate, translate

L = logging.getLogger(__name__)


def bounding_box(obj):
    """Get the (x, y, z) bounding box of an object containing points.

    Returns:
        2D numpy array of [[min_x, min_y, min_z], [max_x, max_y, max_z]]
    """
    return np.array(
        [np.min(obj.points[:, COLS.XYZ], axis=0), np.max(obj.points[:, COLS.XYZ], axis=0)]
    )


def convex_hull(obj):
    """Get the convex hull from object containing points.

    Returns:
        scipy.spatial.ConvexHull object if successful, otherwise None
    """
    return neurom.morphmath.convex_hull(obj.points[:, COLS.XYZ])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Soma classes and functions."""
import math
import warnings

import morphio
import numpy as np
from morphio import SomaError, SomaType

from neurom import morphmath
from neurom.core.dataformat import COLS


class Soma:
    """Base class for a soma.

    Holds a list of raw data rows corresponding to soma points
    and provides iterator access to them.
    """

    def __init__(self, morphio_soma):
        """Constructor.

        Args:
            morphio_soma (morphio.Soma): instance of soma of MorphIO class
        """
        self._morphio_soma = morphio_soma

    def to_morphio(self):
        """Return morphio instance."""
        return self._morphio_soma

    @property
    def center(self):
        """Obtain the center from the first stored point."""
        return get_center(self)

    @property
    def radius(self):
        """Return radius of soma."""
        # this radius is used only for `volume` method, please avoid using it for anything else.
        return get_radius(self)

    def iter(self):
        """Iterator to soma contents."""
        return iter(self.points)

    @property
    def points(self):
        """Get the set of (x, y, z, r) points this soma."""
        return np.concatenate(
            (self._morphio_soma.points, self._morphio_soma.diameters[:, np.newaxis] / 2.0), axis=1
        )

    @property
    def area(self):
        """Calculate soma area."""
        return get_area(self)

    @property
    def volume(self):
        """Calculate soma volume."""
        return get_volume(self)

    def overlaps(self, points, exclude_boundary=False):
        """Check that the given points are located inside the soma."""
        return check_overlaps(self, points, exclude_boundary=exclude_boundary)


class SomaSinglePoint(Soma):
    """Type A: 1point soma.

    Represented by a single point.
    """

    def __str__(self):
        """Return a string representation."""
        return 'SomaSinglePoint(%s) <center: %s, radius: %s>' % (
            repr(self.points),
            self.center,
            self.radius,
        )


class SomaCylinders(Soma):
    """Soma composed of cylinders (like in SWC).

    points describe the locations of the cylinder start/end points, with
    their respective radii, much like how neurites are described:

        ex::

                             /)
                            / o)
                    ______ /  )
                (|)      ) / /
                ( o )    o )/
                (|)_____ )

      Here we have a 'side-view', with each 'o' representing a point, and the
      radius is the height of a '|' character, and the ')' try and show the
      curvature of the cylinger

    Note: when, as in the case above, the cylinder center points don't lie
    in a line, then the overlap between cylinders isn't taken into account for
    the area calculation
    """

    def __str__(self):
        """Return a string representation."""
        return 'SomaCylinders(%s) <center: %s, virtual radius: %s>' % (
            repr(self.points),
            self.center,
            self.radius,
        )


class SomaNeuromorphoThreePointCylinders(SomaCylinders):
    """NeuroMorpho compatible soma.

    Reference:
        http://neuromorpho.org/SomaFormat.html

        Quote (Tue Feb 28 14:26:56 CET 2017):
        The first point constitutes the center of the soma, with coordinates (xs, ys,
        zs) corresponding to the average of all the points in the single contour. An
        equivalent radius (rs) is computed as the average distance of each point of the
        single contour from this center. The first point is assigned parent ID -1 as is
        standard for the root in the SWC format. The second and third soma points, as
        well as all starting points (roots) of dendritic and axonal arbors have this
        first point as the parent (parent ID 1). The x and z coordinates as well as the
        radius of both the second and third soma points are the same as that of the
        first point. The y coordinates of the second and third points are shifted by -rs
        and +rs, respectively (Figure 2). The surface area of this soma cylinder equals
        the surface area of a sphere of radius rs.
    """

    def __init__(self, morphio_soma):
        """Initialize a SomaNeuromorphoThreePointCylinders object."""
        super().__init__(morphio_soma)
        # X    Y     Z   R    P
        # xs ys      zs rs   -1
        # xs (ys-rs) zs rs    1
        # xs (ys+rs) zs rs    1

        r1, r2, r3 = 0.5 * morphio_soma.diameters

        # make sure the above invariant holds
        assert np.isclose(r1, r2) and np.isclose(r1, r3), 'All radii must be the same'

        if r1 < 1e-5:
            warnings.warn('Zero radius for {}'.format(self))

    def __str__(self):
        """Return a string representation."""
        return 'SomaNeuromorphoThreePointCylinders(%s) <center: %s, radius: %s>' % (
            repr(self.points),
            self.center,
            self.radius,
        )


class SomaSimpleContour(Soma):
    """Type C: multiple points soma.

    Represented by a contour.

    The equivalent radius as the average distance to the center.

    Note: This doesn't currently check to see if the contour is in a plane. Also
    the radii of the points are not taken into account.
    """

    def __str__(self):
        """Return a string representation."""
        return 'SomaSimpleContour(%s) <center: %s, radius: %s>' % (
            repr(self.points),
            self.center,
            self.radius,
        )


def _morphio_soma(soma):
    """Return morphio soma."""
    if isinstance(soma, Soma):
        return soma.to_morphio()
    if isinstance(soma, (morphio.Soma, morphio.mut.Soma)):
        return soma
    raise TypeError(f"Unknown soma type {type(soma)}")


def _dispatch_soma_functions(soma, dispatch_mapping, **kwargs):
    morphio_soma = _morphio_soma(soma)
    soma_algo = dispatch_mapping[morphio_soma.type]
    return soma_algo(morphio_soma, **kwargs)


def get_center(soma):
    """Calculate soma center."""
    dispatch_mapping = {
        SomaType.SOMA_UNDEFINED: _first_point_or_none,
        SomaType.SOMA_SINGLE_POINT: _first_point,
        SomaType.SOMA_CYLINDERS: _first_point,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: _first_point,
        SomaType.SOMA_SIMPLE_CONTOUR: _centroid,
    }
    return _dispatch_soma_functions(soma, dispatch_mapping)


def _first_point(morphio_soma):
    """Return the first point."""
    return morphio_soma.points[0]


def _first_point_or_none(morphio_soma):
    """Return first point if there is at least one point or None otherwise."""
    points = morphio_soma.points
    return points[0] if len(points) > 0 else None


def _centroid(morphio_soma):
    """Return the centroid of the soma points."""
    return np.mean(morphio_soma.points, axis=0)


def get_radius(soma):
    """Calculate soma radius."""
    dispatch_mapping = {
        SomaType.SOMA_UNDEFINED: lambda _: 0,
        SomaType.SOMA_SINGLE_POINT: _soma_single_point_radius,
        SomaType.SOMA_CYLINDERS: _soma_cylinders_radius,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: _soma_three_point_cylinders_radius,
        SomaType.SOMA_SIMPLE_CONTOUR: _soma_simple_contour_radius,
    }
    return _dispatch_soma_functions(soma, dispatch_mapping)


def _soma_single_point_radius(morphio_soma):
    """Return first radius."""
    return 0.5 * morphio_soma.diameters[0]


def _soma_cylinders_radius(morphio_soma):
    """Calculate radius calculated from the cylinder area."""
    points = np.concatenate(
        (morphio_soma.points, 0.5 * morphio_soma.diameters[:, np.newaxis]),
        axis=1,
    )
    area = sum(morphmath.segment_area((p0, p1)) for p0, p1 in zip(points, points[1:]))
    return math.sqrt(area / (4.0 * math.pi))


def _soma_three_point_cylinders_radius(morphio_soma):
    """Calculate three-point-cylinder radius."""
    return math.sqrt(get_area(morphio_soma) / (4.0 * math.pi))


def _soma_simple_contour_radius(morphio_soma):
    """Calculate average contour distance from center of soma."""
    return morphmath.average_points_dist(get_center(morphio_soma), morphio_soma.points)


def get_area(soma):
    """Calculate soma area."""
    dispatch_mapping = {
        SomaType.SOMA_UNDEFINED: _soma_undefined_area,
        SomaType.SOMA_SINGLE_POINT: _soma_single_point_area,
        SomaType.SOMA_CYLINDERS: _soma_cylinders_area,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: _soma_three_point_cylinders_area,
        SomaType.SOMA_SIMPLE_CONTOUR: _soma_single_point_area,
    }
    return _dispatch_soma_functions(soma, dispatch_mapping)


def _soma_single_point_area(morphio_soma):
    """Calculate soma area as a sphere."""
    return 4.0 * math.pi * get_radius(morphio_soma) ** 2


def _soma_undefined_area(morphio_soma):
    """Calculate soma as a sphere."""
    warnings.warn('Approximating soma area by a sphere. {}'.format(morphio_soma))
    return _soma_single_point_area(morphio_soma)


def _soma_cylinders_area(morphio_soma):
    """Calculate soma area as a sum of cylinder surfaces withouts caps."""
    points = np.concatenate(
        (
            morphio_soma.points,
            0.5 * morphio_soma.diameters[:, np.newaxis],
        ),
        axis=1,
    )
    return sum(morphmath.segment_area((p0, p1)) for p0, p1 in zip(points, points[1:]))


def _soma_three_point_cylinders_area(morphio_soma):
    """Calculate soma area as a cylinder comprised of three points with the same radius."""
    r = 0.5 * morphio_soma.diameters[0]
    h = morphmath.point_dist(morphio_soma.points[1], morphio_soma.points[2])
    return 2.0 * math.pi * r * h  # ignores the 'end-caps' of the cylinder


def get_volume(soma):
    """Calculate soma volume."""
    dispatch_mapping = {
        SomaType.SOMA_UNDEFINED: _soma_undefined_volume,
        SomaType.SOMA_SINGLE_POINT: _soma_single_point_volume,
        SomaType.SOMA_CYLINDERS: _soma_cylinders_volume,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: _soma_three_point_cylinders_volume,
        SomaType.SOMA_SIMPLE_CONTOUR: _soma_undefined_volume,
    }
    return _dispatch_soma_functions(soma, dispatch_mapping)


def _soma_single_point_volume(morphio_soma):
    """Calculate soma volume as a sphere."""
    return 4.0 / 3 * math.pi * get_radius(morphio_soma) ** 3


def _soma_undefined_volume(morphio_soma):
    """Calculate soma volume as a sphere."""
    warnings.warn('Approximating soma volume by a sphere. {}'.format(morphio_soma))
    return _soma_single_point_volume(morphio_soma)


def _soma_cylinders_volume(morphio_soma):
    """Calculate soma volume as a sum of cylinder volumes."""
    points = np.concatenate(
        (
            morphio_soma.points,
            0.5 * morphio_soma.diameters[:, np.newaxis],
        ),
        axis=1,
    )
    return sum(morphmath.segment_volume((p0, p1)) for p0, p1 in zip(points, points[1:]))


def _soma_three_point_cylinders_volume(morphio_soma):
    """Calculate soma volume as a cylinder of three points and same radius."""
    return 2.0 * math.pi * get_radius(morphio_soma) ** 3


def check_overlaps(soma, points, exclude_boundary=False):
    """Check if soma overlaps with points."""
    dispatch_mapping = {
        SomaType.SOMA_UNDEFINED: _soma_undefined_overlaps,
        SomaType.SOMA_SINGLE_POINT: _soma_undefined_overlaps,
        SomaType.SOMA_CYLINDERS: _soma_cylinders_overlaps,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: _soma_cylinders_overlaps,
        SomaType.SOMA_SIMPLE_CONTOUR: _soma_simple_contour_overlaps,
    }
    return _dispatch_soma_functions(
        soma,
        dispatch_mapping,
        points=points,
        exclude_boundary=exclude_boundary,
    )


def _soma_undefined_overlaps(morphio_soma, points, exclude_boundary):
    """Check if points overlap with soma approximated as a sphere."""
    points = np.atleast_2d(np.asarray(points, dtype=np.float64))

    center = get_center(morphio_soma)
    radius = get_radius(morphio_soma)

    if exclude_boundary:
        return np.linalg.norm(points - center, axis=1) < radius

    return np.linalg.norm(points - center, axis=1) <= radius


def _soma_cylinders_overlaps(morphio_soma, points, exclude_boundary):
    """Check if points overlap with soma approximated as a collection of cylinders."""
    points = np.atleast_2d(np.asarray(points, dtype=np.float64))

    soma_points = np.concatenate(
        (
            morphio_soma.points,
            0.5 * morphio_soma.diameters[:, np.newaxis],
        ),
        axis=1,
    )

    mask = np.ones(len(points)).astype(bool)
    for p1, p2 in zip(soma_points[:-1], soma_points[1:]):
        vec = p2[COLS.XYZ] - p1[COLS.XYZ]
        vec_norm = np.linalg.norm(vec)
        dot = (points[mask] - p1[COLS.XYZ]).dot(vec) / vec_norm

        cross = np.linalg.norm(np.cross(vec, points[mask]), axis=1) / vec_norm
        dot_clipped = np.clip(dot / vec_norm, a_min=0, a_max=1)
        radii = p1[COLS.R] * (1 - dot_clipped) + p2[COLS.R] * dot_clipped

        if exclude_boundary:
            in_cylinder = (dot > 0) & (dot < vec_norm) & (cross < radii)
        else:
            in_cylinder = (dot >= 0) & (dot <= vec_norm) & (cross <= radii)
        mask[np.where(mask)] = ~in_cylinder
        if not mask.any():
            break

    return ~mask


def _soma_simple_contour_overlaps(morphio_soma, points, exclude_boundary):
    """Check if points overlap with soma approximated as a contour."""
    # pylint: disable=too-many-locals
    soma_points = np.concatenate(
        (
            morphio_soma.points,
            0.5 * morphio_soma.diameters[:, np.newaxis],
        ),
        axis=1,
    )
    center = get_center(morphio_soma)

    points = np.atleast_2d(np.asarray(points, dtype=np.float64))

    # Convert points to angles from the center
    relative_pts = points - center
    pt_angles = np.arctan2(relative_pts[:, COLS.Y], relative_pts[:, COLS.X])

    # Convert soma points to angles from the center
    relative_soma_pts = soma_points[:, COLS.XYZ] - center
    soma_angles = np.arctan2(relative_soma_pts[:, COLS.Y], relative_soma_pts[:, COLS.X])

    # Order the soma points by ascending angles
    soma_angle_order = np.argsort(soma_angles)
    ordered_soma_angles = soma_angles[soma_angle_order]
    ordered_relative_soma_pts = relative_soma_pts[soma_angle_order]

    # Find the two soma points which form the segment crossed by the one from the center
    # to the point
    angles = np.atleast_2d(pt_angles).T - ordered_soma_angles
    closest_indices = np.argmin(np.abs(angles), axis=1)
    neighbors = np.ones_like(closest_indices)
    neighbors[angles[np.arange(len(closest_indices)), closest_indices] < 0] = -1
    signs = (neighbors == 1) * 2.0 - 1.0
    neighbors[(closest_indices >= len(relative_soma_pts) - 1) & (neighbors == 1)] = (
        -len(relative_soma_pts) + 1
    )

    # Compute the cross product and multiply by neighbors to get the same result as if all
    # vectors were clockwise
    cross_z = (
        np.cross(
            (
                ordered_relative_soma_pts[closest_indices + neighbors]
                - ordered_relative_soma_pts[closest_indices]
            ),
            relative_pts - ordered_relative_soma_pts[closest_indices],
        )[:, COLS.Z]
        * signs
    )

    if exclude_boundary:
        interior_side = cross_z > 0
    else:
        interior_side = cross_z >= 0

    return interior_side


def make_soma(morphio_soma):
    """Make a soma object from a MorphIO soma.

    Args:
        morphio_soma(morphio.Soma): soma instance of MorphIO
    """
    soma_builders = {
        SomaType.SOMA_SINGLE_POINT: SomaSinglePoint,
        SomaType.SOMA_CYLINDERS: SomaCylinders,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: SomaNeuromorphoThreePointCylinders,
        SomaType.SOMA_SIMPLE_CONTOUR: SomaSimpleContour,
        SomaType.SOMA_UNDEFINED: Soma,
    }

    builder = soma_builders.get(morphio_soma.type)
    if builder is None:
        raise SomaError(f'No NeuroM constructor for MorphIO soma type: {morphio_soma.type}')
    return builder(morphio_soma)



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Advanced analysis examples

These examples highlight more advanced neurom
morphometrics functionality using iterators.

"""
from pathlib import Path

from neurom.core.dataformat import COLS
import neurom as nm
from neurom import geom
from neurom.features import section
from neurom.core import Section
from neurom.core.types import tree_type_checker, NEURITES
from neurom import morphmath as mm
import numpy as np

PACKAGE_DIR = Path(__file__).resolve().parent.parent


def main():
    filename = Path(PACKAGE_DIR, "tests/data/swc/Neuron.swc")

    #  load a neuron from an SWC file
    m = nm.load_morphology(filename)

    # Some examples of what can be done using iteration
    # instead of pre-packaged functions that return lists.
    # The iterations give us a lot of flexibility: we can map
    # any function that takes a segment or section.

    # Get of all neurites in cell by iterating over sections,
    # and summing the section lengths
    def sec_len(sec):
        """Return the length of a section."""
        return mm.section_length(sec.points)

    print("Total neurite length (sections):", sum(sec_len(s) for s in nm.iter_sections(m)))

    # Get length of all neurites in cell by iterating over segments,
    # and summing the segment lengths.
    # This should yield the same result as iterating over sections.
    print(
        "Total neurite length (segments):", sum(mm.segment_length(s) for s in nm.iter_segments(m))
    )

    # get volume of all neurites in cell by summing over segment
    # volumes
    print("Total neurite volume:", sum(mm.segment_volume(s) for s in nm.iter_segments(m)))

    # get area of all neurites in cell by summing over segment
    # areas
    print("Total neurite surface area:", sum(mm.segment_area(s) for s in nm.iter_segments(m)))

    # get total number of neurite points in cell.
    def n_points(sec):
        """number of points in a section."""
        n = len(sec.points)
        # Non-root sections have duplicate first point
        return n if sec.parent is None else n - 1

    print("Total number of points:", sum(n_points(s) for s in nm.iter_sections(m)))

    # get mean radius of neurite points in cell.
    # p[COLS.R] yields the radius for point p.
    # Note: this includes duplicated points at beginning of
    # non-trunk sections
    print("Mean radius of points:", np.mean([s.points[:, COLS.R] for s in nm.iter_sections(m)]))

    # get mean radius of neurite points in cell.
    # p[COLS.R] yields the radius for point p.
    # Note: this includes duplicated points at beginning of
    # non-trunk sections
    pts = [p[COLS.R] for s in m.sections[1:] for p in s.points]
    print("Mean radius of points:", np.mean(pts))

    # get mean radius of segments
    print(
        "Mean radius of segments:", np.mean(list(mm.segment_radius(s) for s in nm.iter_segments(m)))
    )

    # get stats for the segment taper rate, for different types of neurite
    for ttype in NEURITES:
        ttt = ttype
        seg_taper_rate = [
            mm.segment_taper_rate(s)
            for s in nm.iter_segments(m, neurite_filter=tree_type_checker(ttt))
        ]

        print(
            "Segment taper rate (",
            ttype,
            "):\n  mean=",
            np.mean(seg_taper_rate),
            ", std=",
            np.std(seg_taper_rate),
            ", min=",
            np.min(seg_taper_rate),
            ", max=",
            np.max(seg_taper_rate),
            sep="",
        )

    # Number of bifurcation points.
    print(
        "Number of bifurcation points:",
        sum(1 for _ in nm.iter_sections(m, iterator_type=Section.ibifurcation_point)),
    )

    # Number of bifurcation points for apical dendrites
    print(
        "Number of bifurcation points (apical dendrites):",
        sum(
            1
            for _ in nm.iter_sections(
                m,
                iterator_type=Section.ibifurcation_point,
                neurite_filter=tree_type_checker(nm.APICAL_DENDRITE),
            )
        ),
    )

    # Maximum branch order
    print("Maximum branch order:", max(section.branch_order(s) for s in nm.iter_sections(m)))

    # Morphology's bounding box
    # Note: does not account for soma radius
    print("Bounding box ((min x, y, z), (max x, y, z))", geom.bounding_box(m))


if __name__ == "__main__":
    main()



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Anomaly and artefact detection and annotation generation."""
import logging

from neurom.core.dataformat import COLS

L = logging.getLogger(__name__)


def generate_annotation(result, settings):
    """Generate the annotation for a given checker.

    Arguments:
        result (CheckResult): the result of the checker
        settings (dict): the display settings for NeuroLucida

    Returns:
        An S-expression-like string representing the annotation
    """
    if result.status:
        return ''

    header = (
        '\n\n'
        f'({settings["label"]}   ; MUK_ANNOTATION\n'
        f'    (Color {settings["color"]})   ; MUK_ANNOTATION\n'
        f'    (Name "{settings["name"]}")   ; MUK_ANNOTATION'
    )
    points = [p for _, _points in result.info for p in _points]
    annotations = '\n'.join(
        (
            f'    '
            f'({p[COLS.X]:10.2f} {p[COLS.Y]:10.2f} {p[COLS.Z]:10.2f} 0.50)'
            f'   ; MUK_ANNOTATION'
            for p in points
        )
    )
    footer = ')   ; MUK_ANNOTATION\n'

    return f'{header}\n{annotations}\n{footer}'


def annotate(results, settings):
    """Concatenate the annotations of all checkers."""
    annotations = (
        generate_annotation(result, setting) for result, setting in zip(results, settings)
    )
    return '\n'.join(annot for annot in annotations if annot)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Type enumerations."""
import collections.abc
from enum import Enum, EnumMeta, unique

from morphio import SectionType

from neurom.utils import OrderedEnum

_SOMA_SUBTYPE = 31
_ALL_SUBTYPE = 32


@unique
class NeuriteIter(OrderedEnum):
    """Neurite iteration orders."""

    FileOrder = 1  # Order in which neurites appear in the file

    # NRN simulator order: soma -> axon -> basal -> apical
    # Coming from:
    # https://github.com/neuronsimulator/nrn/blob/2dbf2ebf95f1f8e5a9f0565272c18b1c87b2e54c/share/lib/hoc/import3d/import3d_gui.hoc#L874
    NRN = 2


def is_composite_type(subtype):
    """Check that the given type is composite."""
    return NeuriteType(subtype).is_composite()


def _is_sequence(obj):
    return isinstance(obj, collections.abc.Sequence) and not isinstance(obj, str)


def _int_or_tuple(values):
    if isinstance(values, Enum):
        return _int_or_tuple(values.value)

    if isinstance(values, (int, SectionType)):
        return int(values)

    if _is_sequence(values):
        if len(values) == 1:
            return _int_or_tuple(values[0])
        return tuple(_int_or_tuple(v) for v in values)

    raise ValueError(f"Could not cast {values} to int or tuple of ints.")


# pylint: disable=redefined-builtin
class _ArgsIntsOrTuples(EnumMeta):
    def __call__(cls, value, names=None, *, module=None, qualname=None, type=None, start=1):
        try:
            value = _int_or_tuple(value)
        except ValueError:
            pass
        kwargs = {}
        if names is not None:
            # Keep default value of EnumMeta for Python>=3.12.3
            kwargs["names"] = names  # pragma: no cover
        return super().__call__(
            value, module=module, qualname=qualname, type=type, start=start, **kwargs
        )


def _create_neurite_type(cls, value, name=None):
    """Construct and return a cls type."""
    obj = object.__new__(cls)

    # this is an optimization to avoid checks during runtime
    if _is_sequence(value):
        subtypes = value
        root_type = value[0]
    else:
        subtypes = (value,)
        root_type = value

    setattr(obj, "_value_", value)

    if name:
        setattr(obj, "_name_", name)

    obj.subtypes = subtypes
    obj.root_type = root_type

    return obj


# for backward compatibility with 'v1' version
class NeuriteType(Enum, metaclass=_ArgsIntsOrTuples):
    """Type of neurite."""

    axon = SectionType.axon
    apical_dendrite = SectionType.apical_dendrite
    basal_dendrite = SectionType.basal_dendrite
    undefined = SectionType.undefined
    soma = SectionType.soma
    all = SectionType.all
    custom5 = SectionType.custom5
    custom6 = SectionType.custom6
    custom7 = SectionType.custom7
    custom8 = SectionType.custom8
    custom9 = SectionType.custom9
    custom10 = SectionType.custom10
    custom11 = SectionType.custom11
    custom12 = SectionType.custom12
    custom13 = SectionType.custom13
    custom14 = SectionType.custom14
    custom15 = SectionType.custom15
    custom16 = SectionType.custom16
    custom17 = SectionType.custom17
    custom18 = SectionType.custom18
    custom19 = SectionType.custom19

    axon_carrying_dendrite = SectionType.basal_dendrite, SectionType.axon

    def __new__(cls, *values):
        """Construct a NeuriteType from class definitions."""
        return _create_neurite_type(cls, value=_int_or_tuple(values))

    def __hash__(self):
        """Return the has of the type."""
        return hash(self._value_)

    def is_composite(self):
        """Return True if the type consists of more than 1 subtypes."""
        return len(self.subtypes) > 1

    def __eq__(self, other):
        """Equal operator."""
        if not isinstance(other, NeuriteType):
            try:
                other = NeuriteType(other)
            except ValueError:
                return False

        if self.is_composite():
            if other.is_composite():
                is_eq = self.subtypes == other.subtypes
            else:
                is_eq = other.root_type in self.subtypes
        else:
            if other.is_composite():
                is_eq = self.root_type in other.subtypes
            else:
                is_eq = self.root_type == other.root_type
        return is_eq


#: Collection of all neurite types
NEURITES = (NeuriteType.axon, NeuriteType.apical_dendrite, NeuriteType.basal_dendrite)

ROOT_ID = -1


def tree_type_checker(*ref):
    """Tree type checker functor.

    Args:
        *ref(NeuriteType|tuple): Either a single NeuriteType or a variable list of them or a tuple
        of them.

    Returns:
        Functor that takes a tree, and returns true if that tree matches any of
        NeuriteTypes in ref

    Ex:
        >>> import neurom
        >>> from neurom.core.types import NeuriteType, tree_type_checker
        >>> from neurom.core.morphology import Section, iter_neurites
        >>> m = neurom.load_morphology("tests/data/swc/Neuron.swc")
        >>>
        >>> tree_filter = tree_type_checker(NeuriteType.axon, NeuriteType.basal_dendrite)
        >>> it = iter_neurites(m, filt=tree_filter)
        >>>
        >>> tree_filter = tree_type_checker((NeuriteType.axon, NeuriteType.basal_dendrite))
        >>> it = iter_neurites(m, filt=tree_filter)
    """
    ref = tuple(ref)
    if len(ref) == 1 and isinstance(ref[0], (list, tuple)):
        # if `ref` is passed as a tuple of types
        ref = ref[0]
    # validate that all values are of NeuriteType
    ref = [NeuriteType(t) for t in ref]
    if NeuriteType.all in ref:

        def check_tree_type(_):
            """Always returns true."""
            return True

    else:

        def check_tree_type(tree):
            """Check whether tree has the same type as ref.

            Returns:
                True if ref in the same type as tree.type or ref is NeuriteType.all
            """
            return tree.type in ref

    check_tree_type.type = ref

    return check_tree_type


def dendrite_filter(n):
    """Select only dendrites."""
    # pylint: disable=consider-using-in
    return n.type == NeuriteType.basal_dendrite or n.type == NeuriteType.apical_dendrite


def axon_filter(n):
    """Select only axons."""
    return n.type == NeuriteType.axon



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Example for generating density plots."""
from pathlib import Path

import pylab as plt
import matplotlib as mpl
import numpy as np

from neurom import get as get_feat
from neurom.view import matplotlib_utils, matplotlib_impl
from neurom.core.types import NeuriteType

from neurom import load_morphologies

PACKAGE_DIR = Path(__file__).resolve().parent.parent


def extract_density(population, plane="xy", bins=100, neurite_type=NeuriteType.basal_dendrite):
    """Extracts the 2d histogram of the center
    coordinates of segments in the selected plane.
    """
    segment_midpoints = np.array(
        get_feat("segment_midpoints", population, neurite_type=neurite_type)
    )
    horiz = segment_midpoints[:, "xyz".index(plane[0])]
    vert = segment_midpoints[:, "xyz".index(plane[1])]
    return np.histogram2d(np.array(horiz), np.array(vert), bins=(bins, bins))


def plot_density(
    population,  # pylint: disable=too-many-arguments, too-many-locals
    bins=100,
    new_fig=True,
    subplot=111,
    levels=None,
    plane="xy",
    colorlabel="Nodes per unit area",
    labelfontsize=16,
    color_map="Reds",
    no_colorbar=False,
    threshold=0.01,
    neurite_type=NeuriteType.basal_dendrite,
    **kwargs,
):
    """Plots the 2d histogram of the center
    coordinates of segments in the selected plane.
    """
    fig, ax = matplotlib_utils.get_figure(new_fig=new_fig, subplot=subplot)

    H1, xedges1, yedges1 = extract_density(
        population, plane=plane, bins=bins, neurite_type=neurite_type
    )

    mask = H1 < threshold  # mask = H1==0
    H2 = np.ma.masked_array(H1, mask)

    colormap = plt.get_cmap(color_map).copy()
    colormap.set_bad(color='white', alpha=None)

    plots = ax.contourf(
        (xedges1[:-1] + xedges1[1:]) / 2,
        (yedges1[:-1] + yedges1[1:]) / 2,
        np.transpose(H2),  # / np.max(H2),
        cmap=colormap,
        levels=levels,
    )

    if not no_colorbar:
        cbar = plt.colorbar(plots)
        cbar.ax.set_ylabel(colorlabel, fontsize=labelfontsize)

    kwargs["title"] = kwargs.get("title", "")
    kwargs["xlabel"] = kwargs.get("xlabel", plane[0])
    kwargs["ylabel"] = kwargs.get("ylabel", plane[1])

    return matplotlib_utils.plot_style(fig=fig, ax=ax, **kwargs)


def plot_neuron_on_density(
    population,  # pylint: disable=too-many-arguments
    bins=100,
    new_fig=True,
    subplot=111,
    levels=None,
    plane="xy",
    colorlabel="Nodes per unit area",
    labelfontsize=16,
    color_map="Reds",
    no_colorbar=False,
    threshold=0.01,
    neurite_type=NeuriteType.basal_dendrite,
    **kwargs,
):
    """Plots the 2d histogram of the center
    coordinates of segments in the selected plane
    and superimposes the view of the first neurite of the collection.
    """
    _, ax = matplotlib_utils.get_figure(new_fig=new_fig)

    ref_neuron = population[0]
    matplotlib_impl.plot_tree(ref_neuron.neurites[0], ax)

    return plot_density(
        population,
        plane=plane,
        bins=bins,
        new_fig=False,
        subplot=subplot,
        colorlabel=colorlabel,
        labelfontsize=labelfontsize,
        levels=levels,
        color_map=color_map,
        no_colorbar=no_colorbar,
        threshold=threshold,
        neurite_type=neurite_type,
        **kwargs,
    )


def main():
    morphology_directory = Path(PACKAGE_DIR, "tests/data/valid_set")
    neurons = load_morphologies(morphology_directory)

    plot_density(neurons)
    plot_neuron_on_density(neurons)


if __name__ == "__main__":
    main()



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-
#
# NeuroM documentation build configuration file, created by
# sphinx-quickstart on Wed Apr 29 15:13:30 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import importlib.metadata

VERSION = importlib.metadata.version('neurom')

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# General information about the project.
project = u'NeuroM'
author = u'BBP Algorithm Development Section'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = importlib.metadata.version('neurom')
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

suppress_warnings = ["ref.python"]
autosummary_generate = True
autoclass_content = 'both'
autodoc_default_options = {
    'members': True,
}
autosummary_mock_imports = ['plotly']

# autoclass_content = 'both'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx-bluebrain-theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "repo_url": "https://github.com/BlueBrain/NeuroM/",
    "repo_name": "BlueBrain/NeuroM",
}


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'NeuroM'

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = 'scorer.js'

# -- Options for LaTeX output ---------------------------------------------

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# If true, show URL addresses after external links.
# man_show_urls = False


# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Population features.

Any public function from this namespace can be called via features mechanism. Functions in this
namespace can only accept a morphology population as its input no matter how called.

>>> import neurom
>>> from neurom import features
>>> pop = neurom.load_morphologies("tests/data/valid_set")
>>> frequencies = features.get('sholl_frequency', pop)

For more details see :ref:`features`.
"""


from functools import partial

import numpy as np

from neurom.core.dataformat import COLS
from neurom.core.morphology import iter_sections
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as is_type
from neurom.features import NameSpace, feature
from neurom.features import morphology as mf
from neurom.features.morphology import _assert_soma_center

feature = partial(feature, namespace=NameSpace.POPULATION)


@feature(shape=(...,))
def sholl_frequency(morphs, neurite_type=NeuriteType.all, step_size=10, bins=None):
    """Perform Sholl frequency calculations on a population of morphs.

    Args:
        morphs(list|Population): list of morphologies or morphology population
        neurite_type(NeuriteType): which neurites to operate on
        step_size(float): step size between Sholl radii
        bins(iterable of floats): custom binning to use for the Sholl radii. If None, it uses
        intervals of step_size between min and max radii of ``morphs``.
        use_subtrees (bool): Enable mixed subtree processing.

    Note:
        Given a population, the concentric circles range from the smallest soma radius to the
        largest radial neurite distance in steps of `step_size`. Each segment of the morphology is
        tested, so a neurite that bends back on itself, and crosses the same Sholl radius will
        get counted as having crossed multiple times.
    """
    neurite_filter = is_type(neurite_type)

    if bins is None:
        section_iterator = partial(
            iter_sections, neurite_filter=neurite_filter, section_filter=neurite_filter
        )

        max_radius_per_section = [
            np.max(np.linalg.norm(section.points[:, COLS.XYZ] - morph.soma.center, axis=1))
            for morph in map(_assert_soma_center, morphs)
            for section in section_iterator(morph)
        ]

        if not max_radius_per_section:
            return []

        min_soma_edge = min(n.soma.radius for n in morphs)

        bins = np.arange(min_soma_edge, min_soma_edge + max(max_radius_per_section), step_size)

    def _sholl_crossings(morph):
        _assert_soma_center(morph)
        return mf.sholl_crossings(morph, neurite_type, morph.soma.center, bins)

    return np.array([_sholl_crossings(m) for m in morphs]).sum(axis=0).tolist()



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphology classes and functions."""

import warnings
from collections import deque

import morphio
import numpy as np
from cached_property import cached_property

from neurom import morphmath
from neurom.core.dataformat import COLS
from neurom.core.population import Population
from neurom.core.soma import make_soma
from neurom.core.types import NeuriteIter, NeuriteType
from neurom.exceptions import NeuroMError
from neurom.utils import flatten


class Section:
    """Simple recursive tree class."""

    def __init__(self, morphio_section):
        """The section constructor."""
        self._morphio_section = morphio_section

    def to_morphio(self):
        """Returns the morphio section."""
        return self._morphio_section

    @property
    def id(self):
        """Returns the section ID."""
        return self._morphio_section.id

    @property
    def parent(self):
        """Returns the parent section if non root section else None."""
        return None if self.is_root() else Section(self._morphio_section.parent)

    @property
    def children(self):
        """Returns a list of child section."""
        return [Section(child) for child in self._morphio_section.children]

    def is_homogeneous_point(self):
        """A section is homogeneous if it has the same type with its children."""
        return all(c.type == self.type for c in self.children)

    def is_forking_point(self):
        """Is this section a forking point?"""
        return len(self.children) > 1

    def is_bifurcation_point(self):
        """Is tree a bifurcation point?"""
        return len(self.children) == 2

    def is_leaf(self):
        """Is tree a leaf?"""
        return len(self.children) == 0

    def is_root(self):
        """Is tree the root node?"""
        return self._morphio_section.is_root

    def ipreorder(self):
        """Depth-first pre-order iteration of tree nodes."""
        children = deque((self,))
        while children:
            cur_node = children.pop()
            children.extend(reversed(cur_node.children))
            yield cur_node

    def ipostorder(self):
        """Depth-first post-order iteration of tree nodes."""
        children = [
            self,
        ]
        seen = set()
        while children:
            cur_node = children[-1]
            if cur_node not in seen:
                seen.add(cur_node)
                children.extend(reversed(cur_node.children))
            else:
                children.pop()
                yield cur_node

    def iupstream(self, stop_node=None):
        """Iterate from a tree node to the root nodes.

        Args:
            stop_node: Node to stop the upstream traversal. If None, it stops when parent is None.
        """
        if stop_node is None:

            def stop_condition(section):
                return section.is_root()

        else:

            def stop_condition(section):
                return section.is_root() or section == stop_node

        current_section = self
        while not stop_condition(current_section):
            yield current_section
            current_section = current_section.parent
        yield current_section

    def ileaf(self):
        """Iterator to all leaves of a tree."""
        return filter(Section.is_leaf, self.ipreorder())

    def iforking_point(self, iter_mode=ipreorder):
        """Iterator to forking points.

        Args:
            iter_mode: iteration mode. Default: ipreorder.
        """
        return filter(Section.is_forking_point, iter_mode(self))

    def ibifurcation_point(self, iter_mode=ipreorder):
        """Iterator to bifurcation points.

        Args:
            iter_mode: iteration mode. Default: ipreorder.
        """
        return filter(Section.is_bifurcation_point, iter_mode(self))

    def __eq__(self, other):
        """Equal when its morphio section is equal."""
        return self.to_morphio().has_same_shape(other.to_morphio())

    def __hash__(self):
        """Hash of its id."""
        return self.id

    @property
    def segments(self):
        """The array of all segments of the neurite."""
        return list(iter_segments(self))

    @property
    def points(self):
        """Returns the section list of points the NeuroM way (points + radius)."""
        return np.concatenate(
            (self._morphio_section.points, self._morphio_section.diameters[:, np.newaxis] / 2.0),
            axis=1,
        )

    @property
    def type(self):
        """Returns the section type."""
        return NeuriteType(int(self._morphio_section.type))

    @property
    def length(self):
        """Return the path length of this section."""
        return morphmath.section_length(self.points)

    @property
    def area(self):
        """Return the surface area of this section.

        The area is calculated from the segments, as defined by this
        section's points
        """
        return sum(morphmath.segment_area(s) for s in iter_segments(self))

    @property
    def volume(self):
        """Return the volume of this section.

        The volume is calculated from the segments, as defined by this
        section's points
        """
        return sum(morphmath.segment_volume(s) for s in iter_segments(self))

    def __repr__(self):
        """Text representation."""
        parent_id = None if self.parent is None else self.parent.id
        return (
            f'Section(id={self.id}, type={self.type}, n_points={len(self.points)})'
            f'<parent: Section(id={parent_id}), nchildren: {len(self.children)}>'
        )


# NRN simulator iteration order
# See:
# https://github.com/neuronsimulator/nrn/blob/2dbf2ebf95f1f8e5a9f0565272c18b1c87b2e54c/share/lib/hoc/import3d/import3d_gui.hoc#L874
NRN_ORDER = {
    NeuriteType.soma: 0,
    NeuriteType.axon: 1,
    NeuriteType.basal_dendrite: 2,
    NeuriteType.apical_dendrite: 3,
    NeuriteType.undefined: 4,
}


def iter_neurites(obj, mapfun=None, filt=None, neurite_order=NeuriteIter.FileOrder):
    """Iterator to a neurite, morphology or morphology population.

    Applies optional neurite filter and mapping functions.

    Arguments:
        obj: a neurite, morphology or morphology population.
        mapfun: optional neurite mapping function.
        filt: optional neurite filter function.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical

    Examples:
        Get the number of points in each neurite in a morphology population

        >>> from neurom.core.morphology import iter_neurites
        >>> from neurom import load_morphologies
        >>> pop = load_morphologies("tests/data/valid_set")
        >>> n_points = [n for n in iter_neurites(pop, lambda x, section_type: len(x.points))]

        Get the number of points in each axon in a morphology population

        >>> import neurom as nm
        >>> from neurom.core.morphology import iter_neurites
        >>> from neurom import load_morphologies
        >>> pop = load_morphologies("tests/data/valid_set")
        >>> filter = lambda n : n.type == nm.AXON
        >>> mapping = lambda n, section_type: len(n.points)
        >>> n_points = [n for n in iter_neurites(pop, mapping, filter)]
    """
    if isinstance(obj, Neurite):
        neurites = (obj,)
    elif hasattr(obj, "neurites"):
        neurites = obj.neurites
    else:
        neurites = obj

    if neurite_order == NeuriteIter.NRN:
        if isinstance(obj, Population):
            warnings.warn(
                '`iter_neurites` with `neurite_order` over Population orders neurites'
                'within the whole population, not within each morphology separately.'
            )
        last_position = max(NRN_ORDER.values()) + 1
        neurites = sorted(neurites, key=lambda neurite: NRN_ORDER.get(neurite.type, last_position))

    neurite_iter = iter(neurites) if filt is None else filter(filt, neurites)

    if mapfun is None:
        return neurite_iter

    return (
        (
            mapfun(
                neurite,
                section_type=filt.type if filt is not None else None,
            )
            if neurite.process_subtrees
            else mapfun(neurite, section_type=NeuriteType.all)
        )
        for neurite in neurite_iter
    )


def iter_sections(
    neurites,
    iterator_type=Section.ipreorder,
    neurite_filter=None,
    neurite_order=NeuriteIter.FileOrder,
    section_filter=None,
):
    """Iterator to the sections in a neurite, morphology or morphology population.

    Arguments:
        neurites: morphology, population, neurite, or iterable containing neurite objects
        iterator_type: section iteration order within a given neurite. Must be one of:
            Section.ipreorder: Depth-first pre-order iteration of tree nodes
            Section.ipostorder: Depth-first post-order iteration of tree nodes
            Section.iupstream: Iterate from a tree node to the root nodes
            Section.ibifurcation_point: Iterator to bifurcation points
            Section.ileaf: Iterator to all leaves of a tree

        neurite_filter: optional top level filter on properties of neurite neurite objects.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical
        section_filter: optional section level filter. Please note that neurite_filter takes
            precedence over the section_filter.


    Examples:
        Get the number of points in each section of all the axons in a morphology population

        >>> import neurom as nm
        >>> from neurom.core.morphology import iter_sections
        >>> filter = lambda n : n.type == nm.AXON
        >>> n_points = [len(s.points) for s in iter_sections(pop,  neurite_filter=filter)]
    """
    neurites = iter_neurites(neurites, filt=neurite_filter, neurite_order=neurite_order)
    sections = flatten(iterator_type(neurite.root_node) for neurite in neurites)
    return sections if section_filter is None else filter(section_filter, sections)


def iter_segments(
    obj,
    neurite_filter=None,
    neurite_order=NeuriteIter.FileOrder,
    section_filter=None,
    section_iterator=Section.ipreorder,
):
    """Return an iterator to the segments in a collection of neurites.

    Arguments:
        obj: morphology, population, neurite, section, or iterable containing neurite objects
        neurite_filter: optional top level filter on properties of neurite neurite objects
        neurite_order: order upon which neurite should be iterated. Values:
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical
        section_filter: optional section level filter
        section_iterator: section iteration order within a given neurite. Must be one of:
            Section.ipreorder: Depth-first pre-order iteration of tree nodes
            Section.ipostorder: Depth-first post-order iteration of tree nodes
            Section.iupstream: Iterate from a tree node to the root nodes
            Section.ibifurcation_point: Iterator to bifurcation points
            Section.ileaf: Iterator to all leaves of a tree

    Note:
        This is a convenience function provided for generic access to
        morphology segments. It may have a performance overhead WRT custom-made
        segment analysis functions that leverage numpy and section-wise iteration.
    """
    sections = iter(
        (obj,)
        if isinstance(obj, Section)
        else iter_sections(
            obj,
            iterator_type=section_iterator,
            neurite_filter=neurite_filter,
            neurite_order=neurite_order,
            section_filter=section_filter,
        )
    )

    return flatten(zip(section.points[:-1], section.points[1:]) for section in sections)


def iter_points(obj, neurite_filter=None, neurite_order=NeuriteIter.FileOrder, section_filter=None):
    """Return an iterator to the points in a population, morphology, neurites, or section.

    Args:
        obj: population, morphology, neurite, section or iterable containing
        neurite_filter: optional top level filter on properties of neurite neurite objects
        neurite_order: order upon which neurite should be iterated. Values:
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical
        section_filter: optional section level filter
    """
    sections = (
        iter((obj,))
        if isinstance(obj, Section)
        else iter_sections(
            obj,
            neurite_filter=neurite_filter,
            neurite_order=neurite_order,
            section_filter=section_filter,
        )
    )

    return flatten(s.points[:, COLS.XYZ] for s in sections)


def graft_morphology(section):
    """Returns a morphology starting at section."""
    assert isinstance(section, Section)
    m = morphio.mut.Morphology()
    m.append_root_section(section.to_morphio())
    return Morphology(m)


class Neurite:
    """Class representing a neurite tree."""

    def __init__(self, root_node, *, process_subtrees=False):
        """Constructor.

        Args:
            root_node (morphio.Section): root section
            process_subtrees (bool): enable mixed tree processing if set to True
        """
        self._root_node = root_node
        self._process_subtrees = process_subtrees

    @property
    def process_subtrees(self):
        """Enable mixed tree processing if set to True."""
        return self._process_subtrees

    @process_subtrees.setter
    def process_subtrees(self, value):
        self._process_subtrees = value
        if "type" in vars(self):
            del vars(self)["type"]

    @property
    def morphio_root_node(self):
        """Returns the morphio root section."""
        return self._root_node

    @property
    def root_node(self):
        """The first section of the neurite."""
        return Section(self.morphio_root_node)

    @cached_property
    def type(self):
        """The type of the Neurite (which can be composite)."""
        return NeuriteType(self.subtree_types)

    @cached_property
    def subtree_types(self):
        """The types of the subtrees."""
        if not self._process_subtrees:
            return NeuriteType(self.morphio_root_node.type)

        it = self.root_node.ipreorder()
        subtree_types = [next(it).to_morphio().type]

        for section in it:
            if section.type != section.parent.type:
                subtree_types.append(NeuriteType(section.to_morphio().type))

        return subtree_types

    @property
    def sections(self):
        """The array of all sections."""
        return list(iter_sections(self))

    @property
    def segments(self):
        """The array of all segments of the neurite."""
        return list(iter_segments(self))

    @property
    def points(self):
        """Array with all the points in this neurite.

        Note: Duplicate points at section bifurcations are removed
        """
        # Neurite first point must be added manually
        _ptr = [self.root_node.points[0][COLS.XYZR]]
        for s in iter_sections(self):
            _ptr.append(s.points[1:, COLS.XYZR])
        return np.vstack(_ptr)

    @property
    def length(self):
        """Returns the total length of this neurite.

        The length is defined as the sum of lengths of the sections.
        """
        # pylint: disable=import-outside-toplevel
        from neurom.features.neurite import total_length

        return total_length(self)

    @property
    def area(self):
        """Return the surface area of this neurite.

        The area is defined as the sum of area of the sections.
        """
        # pylint: disable=import-outside-toplevel
        from neurom.features.neurite import total_area

        return total_area(self)

    @property
    def volume(self):
        """Return the volume of this neurite.

        The volume is defined as the sum of volumes of the sections.
        """
        # pylint: disable=import-outside-toplevel
        from neurom.features.neurite import total_volume

        return total_volume(self)

    def is_heterogeneous(self) -> bool:
        """Returns true if the neurite consists of more that one section types."""
        return self.morphio_root_node.is_heterogeneous()

    def __eq__(self, other):
        """If root node ids and types are equal."""
        return (
            self.type == other.type
            and self.morphio_root_node.id == other.morphio_root_node.id
            and self.process_subtrees == other.process_subtrees
        )

    def __hash__(self):
        """Hash is made of tuple of type and root_node."""
        return hash((self.type, self.root_node, self.process_subtrees))

    def __repr__(self):
        """Return a string representation."""
        return 'Neurite <type: %s>' % self.type


class Morphology:
    """Class representing a simple morphology."""

    def __init__(self, morphio_morph, name=None, *, process_subtrees=False):
        """Morphology constructor.

        Args:
            morphio_morph (morphio.Morphology|morphio.mut.Morphology): a morphio object
            name (str): an optional morphology name
            process_subtrees (bool): enable mixed tree processing if set to True
        """
        if not isinstance(morphio_morph, (morphio.Morphology, morphio.mut.Morphology)):
            raise NeuroMError(
                f"Expected morphio Morphology object but got: {morphio_morph}.\n"
                f"Use neurom.load_morphology() to load from file."
            )

        self._morphio_morph = morphio_morph

        self.name = name if name else 'Morphology'
        self.soma = make_soma(self._morphio_morph.soma)

        self.process_subtrees = process_subtrees

    def to_morphio(self):
        """Returns the morphio morphology object."""
        return self._morphio_morph

    def copy(self):
        """Returns a shallow copy of the morphio morphology object."""
        return Morphology(self.to_morphio(), name=self.name, process_subtrees=self.process_subtrees)

    @property
    def neurites(self):
        """The list of neurites."""
        return [
            Neurite(root_section, process_subtrees=self.process_subtrees)
            for root_section in self._morphio_morph.root_sections
        ]

    def section(self, section_id):
        """Returns the section with the given id."""
        return Section(self._morphio_morph.section(section_id))

    @property
    def sections(self):
        """The array of all sections, excluding the soma."""
        return list(iter_sections(self))

    @property
    def segments(self):
        """The array of all segments of the sections."""
        return list(iter_segments(self))

    @property
    def points(self):
        """Returns the list of points."""
        return np.concatenate([section.points for section in iter_sections(self)])

    def transform(self, trans):
        """Return a copy of this morphology with a 3D transformation applied."""
        morph = self._morphio_morph

        is_immutable = hasattr(morph, 'as_mutable')

        # make copy or convert to mutable if immutable
        if is_immutable:
            morph = morph.as_mutable()
        else:
            morph = morphio.mut.Morphology(morph)

        morph.soma.points = trans(morph.soma.points)

        for section in morph.iter():
            section.points = trans(section.points)

        if is_immutable:
            return Morphology(morph.as_immutable())
        return Morphology(morph)

    def __copy__(self):
        """Creates a deep copy of Morphology instance."""
        return Morphology(self.to_morphio(), self.name)

    def __deepcopy__(self, memodict={}):
        """Creates a deep copy of Morphology instance."""
        # pylint: disable=dangerous-default-value
        return Morphology(self.to_morphio(), self.name)

    def __repr__(self):
        """Return a string representation."""
        return 'Morphology <soma: %s, n_neurites: %d>' % (self.soma, len(self.neurites))



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphology features.

Any public function from this namespace can be called via the features mechanism. If calling
directly the function in this namespace can only accept a morphology as its input. If you want to
apply it to a morphology population then you must use the features mechanism e.g. ``features.get``.
The features mechanism does not allow you to apply these features to neurites.

>>> import neurom
>>> from neurom import features
>>> m = neurom.load_morphology("tests/data/swc/Neuron.swc")
>>> result = features.get('soma_surface_area', m)
>>> population = neurom.load_morphologies("tests/data/valid_set")
>>> result = features.get('sholl_crossings', population)

For more details see :ref:`features`.
"""

import warnings
from collections.abc import Iterable
from functools import partial

import numpy as np

import neurom.core.soma
from neurom import morphmath
from neurom.core.dataformat import COLS
from neurom.core.morphology import (
    Morphology,
    iter_neurites,
    iter_points,
    iter_sections,
    iter_segments,
)
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as is_type
from neurom.exceptions import NeuroMError
from neurom.features import NameSpace, feature
from neurom.features import neurite as nf
from neurom.features import section as sf
from neurom.morphmath import convex_hull
from neurom.utils import flatten, str_to_plane

feature = partial(feature, namespace=NameSpace.NEURON)


def _assert_soma_center(morph):
    if morph.soma.center is None:
        raise NeuroMError(
            f"The morphology named '{morph.name}' has no soma so the feature can not be computed."
        )
    return morph


def _map_neurites(function, morph, neurite_type):
    return list(
        iter_neurites(
            obj=morph,
            mapfun=function,
            filt=is_type(neurite_type),
        )
    )


def _map_neurite_root_nodes(function, morph, neurite_type):
    if neurite_type == NeuriteType.all:
        filt = None
    else:

        def filt(neurite):
            return neurite_type == neurite.type.root_type

    return [function(trunk.root_node) for trunk in iter_neurites(obj=morph, filt=filt)]


def _filter_mode(obj, neurite_type):
    if obj.process_subtrees:
        return {"section_filter": is_type(neurite_type)}
    return {"neurite_filter": is_type(neurite_type)}


def _get_sections(morph, neurite_type):
    return list(iter_sections(morph, **_filter_mode(morph, neurite_type)))


def _get_segments(morph, neurite_type):
    return list(iter_segments(morph, **_filter_mode(morph, neurite_type)))


def _get_points(morph, neurite_type):
    return list(iter_points(morph, **_filter_mode(morph, neurite_type)))


@feature(shape=())
def soma_volume(morph):
    """Get the volume of a morphology's soma."""
    return neurom.core.soma.get_volume(morph.soma)


@feature(shape=())
def soma_surface_area(morph):
    """Get the surface area of a morphology's soma.

    Note:
        The surface area is calculated by assuming the soma is spherical.
    """
    return neurom.core.soma.get_area(morph.soma)


@feature(shape=())
def soma_radius(morph):
    """Get the radius of a morphology's soma."""
    return neurom.core.soma.get_radius(morph.soma)


@feature(shape=())
def max_radial_distance(morph, origin=None, neurite_type=NeuriteType.all):
    """Get the maximum radial distances of the termination sections."""
    origin = morph.soma.center if origin is None else origin

    term_radial_distances = _map_neurites(
        partial(nf.max_radial_distance, origin=origin), morph, neurite_type
    )
    return max(term_radial_distances) if term_radial_distances else 0.0


@feature(shape=(...,))
def section_radial_distances(morph, origin=None, neurite_type=NeuriteType.all):
    """Section radial distances."""
    origin = morph.soma.center if origin is None else origin

    return list(
        flatten(
            _map_neurites(
                partial(nf.section_radial_distances, origin=origin),
                morph=morph,
                neurite_type=neurite_type,
            )
        )
    )


@feature(shape=(...,))
def section_term_radial_distances(morph, origin=None, neurite_type=NeuriteType.all):
    """Get the radial distances of the termination sections."""
    origin = morph.soma.center if origin is None else origin

    return list(
        flatten(
            _map_neurites(
                partial(nf.section_term_radial_distances, origin=origin),
                morph=morph,
                neurite_type=neurite_type,
            )
        )
    )


@feature(shape=(...,))
def section_bif_radial_distances(morph, origin=None, neurite_type=NeuriteType.all):
    """Get the radial distances of the bifurcation sections."""
    origin = morph.soma.center if origin is None else origin

    return list(
        flatten(
            _map_neurites(
                partial(nf.section_bif_radial_distances, origin=origin),
                morph=morph,
                neurite_type=neurite_type,
            )
        )
    )


@feature(shape=(...,))
def segment_radial_distances(morph, origin=None, neurite_type=NeuriteType.all):
    """Ger the radial distances of the segments."""
    origin = morph.soma.center if origin is None else origin

    return list(
        flatten(
            _map_neurites(
                partial(nf.segment_radial_distances, origin=origin),
                morph=morph,
                neurite_type=neurite_type,
            )
        )
    )


@feature(shape=(...,))
def number_of_sections_per_neurite(morph, neurite_type=NeuriteType.all):
    """List of numbers of sections per neurite."""
    return _map_neurites(nf.number_of_sections, morph, neurite_type)


@feature(shape=(...,))
def total_length_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite lengths."""
    return _map_neurites(nf.total_length, morph, neurite_type)


@feature(shape=(...,))
def total_area_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite areas."""
    return _map_neurites(nf.total_area, morph, neurite_type)


@feature(shape=(...,))
def total_volume_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite volumes."""
    return _map_neurites(nf.total_volume, morph, neurite_type)


@feature(shape=(...,))
def trunk_origin_azimuths(morph, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin azimuths of a morph.

    The azimuth is defined as Angle between x-axis and the vector
    defined by (initial tree point - soma center) on the x-z plane.

    The range of the azimuth angle [-pi, pi] radians
    """
    _assert_soma_center(morph)

    def azimuth(root_node):
        """Azimuth of a neurite trunk."""
        return morphmath.azimuth_from_vector(
            morphmath.vector(root_node.points[0], morph.soma.center)
        )

    return _map_neurite_root_nodes(azimuth, morph, neurite_type)


@feature(shape=(...,))
def trunk_origin_elevations(morph, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin elevations of a morph.

    The elevation is defined as the angle between x-axis and the
    vector defined by (initial tree point - soma center)
    on the x-y half-plane.

    The range of the elevation angle [-pi/2, pi/2] radians
    """
    _assert_soma_center(morph)

    def elevation(root_node):
        """Elevation of a section."""
        return morphmath.elevation_from_vector(
            morphmath.vector(root_node.points[0], morph.soma.center)
        )

    return _map_neurite_root_nodes(elevation, morph, neurite_type)


@feature(shape=(...,))
def trunk_vectors(morph, neurite_type=NeuriteType.all):
    """Calculate the vectors between all the trunks of the morphology and the soma center."""
    _assert_soma_center(morph)

    def vector_from_soma_to_root(root_node):
        return morphmath.vector(root_node.points[0], morph.soma.center)

    return _map_neurite_root_nodes(vector_from_soma_to_root, morph, neurite_type)


@feature(shape=(...,))
def trunk_angles(
    morph,
    neurite_type=NeuriteType.all,
    coords_only="xy",
    sort_along="xy",
    consecutive_only=True,
):
    """Calculate the angles between all the trunks of the morph.

    By default, the angles are defined on the x-y plane and the trees are sorted from the y axis
    and anticlock-wise.

    Args:
        morph: The morphology to process.
        neurite_type: Only the neurites of this type are considered.
        coords_only: Consider only the coordinates listed in this argument (should be a combination
            of 'x', 'y' and 'z').
        sort_along: Sort angles according to the given plane (should be 'xy', 'xz' or 'yz') before
            computing the angles between the trunks.
        consecutive_only: Compute only the angles between consecutive trunks (the default order of
            neurite trunks is the same as the one used by
            :func:`neurom.core.morphology.iter_neurites` but this order can be changed using the
            `sort_along` parameter).

    Returns:
        list[[float]] or list[float]:
            The angles between each trunk and all the others. If ``consecutive_only`` is ``True``,
            only the angle with the next trunk is returned for each trunk.
    """
    vectors = np.array(trunk_vectors(morph, neurite_type=neurite_type))
    # In order to avoid the failure of the process in case the neurite_type does not exist
    if len(vectors) == 0:
        return []

    if sort_along:
        # Sorting angles according to the given plane
        sort_coords = str_to_plane(sort_along)
        order = np.argsort(
            np.fromiter(
                (
                    morphmath.angle_between_projections(i / np.linalg.norm(i), [0, 1])
                    for i in vectors[:, sort_coords]
                ),
                dtype=float,
            )
        )
        vectors = vectors[order]

    # Select coordinates to consider
    if coords_only:
        coords = str_to_plane(coords_only)
        vectors = vectors[:, coords]

    # Compute angles between each trunk and the next ones
    n_vectors = len(vectors)
    cycling_vectors = np.vstack([vectors, vectors])
    angles = [
        (
            num_i,
            [
                morphmath.angle_between_vectors(i, j)
                for j in cycling_vectors[num_i : num_i + n_vectors]
            ],
        )
        for num_i, i in enumerate(vectors)
    ]

    if consecutive_only:
        angles = [i[1][-1] for i in angles if i[1]]
    else:
        angles = [i[1] for i in angles]

    return angles


@feature(shape=(...,))
def trunk_angles_inter_types(
    morph,
    source_neurite_type=NeuriteType.apical_dendrite,
    target_neurite_type=NeuriteType.basal_dendrite,
    closest_component=None,
):
    """Calculate the angles between the trunks of the morph of a source type to target type.

    Args:
        morph: The morphology to process.
        source_neurite_type: Only the neurites of this type are considered as sources.
        target_neurite_type: Only the neurites of this type are considered as targets.
        closest_component:
            If ``closest_component`` is not ``None``, only one element is returned for each neurite
            of source type:

            * if set to 0, the one with the lowest absolute 3d angle is returned.
            * if set to 1, the one with the lowest absolute elevation angle is returned.
            * if set to 2, the one with the lowest absolute azimuth angle is returned.

    Returns:
        list[list[float]] or list[float]:
            If ``closest_component`` is ``None``, a list of 3 elements is returned for each couple
            of neurites:

            * the absolute 3d angle between the two vectors.
            * the elevation angle (or polar angle) between the two vectors.
            * the azimuth angle between the two vectors.

            If ``closest_component`` is not ``None``, only one of these values is returned for each
            couple.
    """
    source_vectors = trunk_vectors(morph, neurite_type=source_neurite_type)
    target_vectors = trunk_vectors(morph, neurite_type=target_neurite_type)

    # In order to avoid the failure of the process in case the neurite_type does not exist
    if len(source_vectors) == 0 or len(target_vectors) == 0:
        return []

    angles = np.empty((len(source_vectors), len(target_vectors), 3), dtype=float)

    for i, source in enumerate(source_vectors):
        for j, target in enumerate(target_vectors):
            angles[i, j, 0] = morphmath.angle_between_vectors(source, target)
            angles[i, j, [1, 2]] = morphmath.spherical_from_vector(
                target
            ) - morphmath.spherical_from_vector(source)

    # Ensure elevation differences are in [-pi, pi]
    angles[:, :, 1] = morphmath.angles_to_pi_interval(angles[:, :, 1])

    # Ensure azimuth differences are in [-2pi, 2pi]
    angles[:, :, 2] = morphmath.angles_to_pi_interval(angles[:, :, 2], scale=2.0)

    if closest_component is not None:
        angles = angles[
            np.arange(len(angles)), np.argmin(np.abs(angles[:, :, closest_component]), axis=1)
        ][:, np.newaxis, :]

    return angles.tolist()


@feature(shape=(...,))
def trunk_angles_from_vector(
    morph,
    neurite_type=NeuriteType.all,
    vector=None,
):
    """Calculate the angles between the trunks of the morph of a given type and a given vector.

    Args:
        morph: The morphology to process.
        neurite_type: Only the neurites of this type are considered.
        vector: The reference vector. If ``None``, the reference vector is set to ``(0, 1, 0)``.

    Returns:
        list[list[float]]:
            For each neurite, an array with 3 elements is returned:

            * the absolute 3d angle between the two vectors.
            * the elevation angle (or polar angle) between the two vectors.
            * the azimuth angle between the two vectors.
    """
    if vector is None:
        vector = (0, 1, 0)

    vectors = np.array(trunk_vectors(morph, neurite_type=neurite_type))

    # In order to avoid the failure of the process in case the neurite_type does not exist
    if len(vectors) == 0:
        return []

    angles = np.empty((len(vectors), 3), dtype=float)
    for i, i_vec in enumerate(vectors):
        angles[i, 0] = morphmath.angle_between_vectors(vector, i_vec)
        angles[i, (1, 2)] = morphmath.spherical_from_vector(
            i_vec
        ) - morphmath.spherical_from_vector(vector)

    # Ensure elevation difference are in [-pi, pi]
    angles[:, 1] = morphmath.angles_to_pi_interval(angles[:, 1])

    # Ensure azimuth difference are in [-2pi, 2pi]
    angles[:, 2] = morphmath.angles_to_pi_interval(angles[:, 2], scale=2)

    return angles.tolist()


@feature(shape=(...,))
def trunk_origin_radii(
    morph,
    neurite_type=NeuriteType.all,
    min_length_filter=None,
    max_length_filter=None,
):
    """Radii of the trunk sections of neurites in a morph.

    .. warning::
        If ``min_length_filter`` and / or ``max_length_filter`` is given, the points are filtered
        and the mean radii of the remaining points is returned.
        Note that if the ``min_length_filter`` is greater than the path distance of the last point
        of the first section, the radius of this last point is returned.

    Args:
        morph: The morphology to process.
        neurite_type: Only the neurites of this type are considered.
        min_length_filter: The min length from which the neurite points are considered.
        max_length_filter: The max length from which the neurite points are considered.

    Returns:
        list[float]:
            * if ``min_length_filter`` and ``max_length_filter`` are ``None``, the radii of the
              first point of each neurite are returned.
            * else the mean radius of the points between the given ``min_length_filter`` and
              ``max_length_filter`` are returned.
    """
    if min_length_filter is not None and min_length_filter <= 0:
        raise NeuroMError(
            "In 'trunk_origin_radii': the 'min_length_filter' value must be strictly greater "
            "than 0."
        )

    if max_length_filter is not None and max_length_filter <= 0:
        raise NeuroMError(
            "In 'trunk_origin_radii': the 'max_length_filter' value must be strictly greater "
            "than 0."
        )

    if (
        min_length_filter is not None
        and max_length_filter is not None
        and min_length_filter >= max_length_filter
    ):
        raise NeuroMError(
            "In 'trunk_origin_radii': the 'min_length_filter' value must be strictly less than the "
            "'max_length_filter' value."
        )

    def trunk_first_radius(root_node):
        return root_node.points[0][COLS.R]

    def trunk_mean_radius(root_node):
        points = root_node.points

        interval_lengths = morphmath.interval_lengths(points)
        path_lengths = np.insert(np.cumsum(interval_lengths), 0, 0)
        valid_pts = np.ones(len(path_lengths), dtype=bool)

        if min_length_filter is not None:
            valid_pts = valid_pts & (path_lengths >= min_length_filter)
            if not valid_pts.any():
                warnings.warn(
                    "In 'trunk_origin_radii': the 'min_length_filter' value is greater than the "
                    "path distance of the last point of the last section so the radius of this "
                    "point is returned."
                )
                return points[-1, COLS.R]

        if max_length_filter is not None:
            valid_max = path_lengths <= max_length_filter
            valid_pts = valid_pts & valid_max
            if not valid_pts.any():
                warnings.warn(
                    "In 'trunk_origin_radii': the 'min_length_filter' and 'max_length_filter' "
                    "values excluded all the points of the section so the radius of the first "
                    "point after the 'min_length_filter' path distance is returned."
                )
                # pylint: disable=invalid-unary-operand-type
                return points[~valid_max, COLS.R][0]

        return points[valid_pts, COLS.R].mean()

    function = (
        trunk_first_radius
        if max_length_filter is None and min_length_filter is None
        else trunk_mean_radius
    )

    return _map_neurite_root_nodes(function, morph, neurite_type)


@feature(shape=(...,))
def trunk_section_lengths(morph, neurite_type=NeuriteType.all):
    """List of lengths of trunk sections of neurites in a morph."""
    return _map_neurite_root_nodes(sf.section_length, morph, neurite_type)


@feature(shape=())
def number_of_neurites(morph, neurite_type=NeuriteType.all):
    """Number of neurites in a morph."""
    return len(_map_neurites(lambda x, section_type: 1, morph, neurite_type))


@feature(shape=(...,))
def neurite_volume_density(morph, neurite_type=NeuriteType.all):
    """Get volume density per neurite."""
    return _map_neurites(nf.volume_density, morph, neurite_type)


@feature(shape=(...,))
def sholl_crossings(morph, neurite_type=NeuriteType.all, center=None, radii=None):
    """Calculate crossings of neurites.

    Args:
        morph(Morphology|list): morphology or a list of neurites
        neurite_type(NeuriteType): Type of neurite to use. By default ``NeuriteType.all`` is used.
        center(Point): center point, if None then soma center is taken
        radii(iterable of floats): radii for which crossings will be counted,
            if None then soma radius is taken

    Returns:
        Array of same length as radii, with a count of the number of crossings
        for the respective radius

    This function can also be used with a list of sections, as follow::

        secs = (sec for sec in nm.iter_sections(morph) if complex_filter(sec))
        sholl = nm.features.neuritefunc.sholl_crossings(secs,
                                                        center=morph.soma.center,
                                                        radii=np.arange(0, 1000, 100))
    """

    def count_crossings(section, radius):
        """Used to count crossings of segments in neurite with radius."""
        r2 = radius**2
        count = 0
        for start, end in iter_segments(section):
            start_dist2, end_dist2 = (
                morphmath.point_dist2(center, start),
                morphmath.point_dist2(center, end),
            )

            if start_dist2 <= r2 <= end_dist2 or end_dist2 <= r2 <= start_dist2:
                count += 1

        return count

    if center is None or radii is None:
        assert isinstance(morph, Morphology) and morph.soma, (
            '`sholl_crossings` input error. If `center` or `radii` is not set then `morph` is '
            'expected to be an instance of Morphology and have a soma.'
        )
        if center is None:
            _assert_soma_center(morph)
            center = morph.soma.center
        if radii is None:
            radii = [morph.soma.radius]

    if isinstance(morph, Iterable):
        sections = filter(is_type(neurite_type), morph)
    else:
        sections = _get_sections(morph, neurite_type)

    counts_per_radius = [0 for _ in range(len(radii))]

    for section in sections:
        for i, radius in enumerate(radii):
            counts_per_radius[i] += count_crossings(section, radius)

    return counts_per_radius


@feature(shape=(...,))
def sholl_frequency(morph, neurite_type=NeuriteType.all, step_size=10, bins=None):
    """Perform Sholl frequency calculations on a morph.

    Args:
        morph(Morphology): a morphology
        neurite_type(NeuriteType): which neurites to operate on
        step_size(float): step size between Sholl radii
        bins(iterable of floats): custom binning to use for the Sholl radii. If None, it uses
        intervals of step_size between min and max radii of ``morphologies``.

    Note:
        Given a morphology, the soma center is used for the concentric circles,
        which range from the soma radii, and the maximum radial distance
        in steps of `step_size`. Each segment of the morphology is tested, so a neurite that
        bends back on itself, and crosses the same Sholl radius will get counted as
        having crossed multiple times.

        If a `neurite_type` is specified and there are no trees corresponding to it, an empty
        list will be returned.
    """
    _assert_soma_center(morph)

    if bins is None:
        min_soma_edge = morph.soma.radius

        sections = _get_sections(morph, neurite_type)

        max_radius_per_section = [
            np.max(np.linalg.norm(section.points[:, COLS.XYZ] - morph.soma.center, axis=1))
            for section in sections
        ]

        if not max_radius_per_section:
            return []

        bins = np.arange(min_soma_edge, min_soma_edge + max(max_radius_per_section), step_size)

    return sholl_crossings(morph, neurite_type, morph.soma.center, bins)


def _extent_along_axis(morph, axis, neurite_type):
    """Returns the total extent of the morpholog neurites.

    The morphology is filtered by neurite type and the extent is calculated
    along the coordinate axis direction (e.g. COLS.X).
    """
    points = _get_points(morph, neurite_type)

    if not points:
        return 0.0

    return abs(np.ptp(np.asarray(points)[:, axis]))


@feature(shape=())
def total_width(morph, neurite_type=NeuriteType.all):
    """Extent of morphology along axis x."""
    return _extent_along_axis(morph, COLS.X, neurite_type)


@feature(shape=())
def total_height(morph, neurite_type=NeuriteType.all):
    """Extent of morphology along axis y."""
    return _extent_along_axis(morph, COLS.Y, neurite_type)


@feature(shape=())
def total_depth(morph, neurite_type=NeuriteType.all):
    """Extent of morphology along axis z."""
    return _extent_along_axis(morph, COLS.Z, neurite_type)


@feature(shape=())
def volume_density(morph, neurite_type=NeuriteType.all):
    """Get the volume density.

    The volume density is defined as the ratio of the neurite volume and
    the volume of the morphology's enclosing convex hull

    .. note:: Returns `np.nan` if the convex hull computation fails or there are not points
              available due to neurite type filtering.
    """
    points = _get_points(morph, neurite_type)

    if not points:
        return np.nan

    morph_hull = convex_hull(points)

    if morph_hull is None:
        return np.nan

    total_volume = sum(total_volume_per_neurite(morph, neurite_type=neurite_type))

    return total_volume / morph_hull.volume


def _unique_projected_points(morph, projection_plane, neurite_type):
    key = "".join(sorted(projection_plane.lower()))

    try:
        axes = {"xy": COLS.XY, "xz": COLS.XZ, "yz": COLS.YZ}[key]

    except KeyError as e:
        raise NeuroMError(
            f"Invalid 'projection_plane' argument {projection_plane}. "
            f"Please select 'xy', 'xz', or 'yz'."
        ) from e

    points = _get_points(morph, neurite_type)

    if len(points) == 0:
        return np.empty(shape=(0, 3), dtype=np.float32)

    return np.unique(np.vstack(points), axis=0)[:, axes]


@feature(shape=())
def aspect_ratio(morph, neurite_type=NeuriteType.all, projection_plane="xy"):
    """Calculates the min/max ratio of the principal direction extents along the plane.

    Args:
        morph: Morphology object.
        neurite_type: The neurite type to use. By default all neurite types are used.
        projection_plane: Projection plane to use for the calculation. One of ('xy', 'xz', 'yz').

    Returns:
        The aspect ratio feature of the morphology points.
    """
    projected_points = _unique_projected_points(morph, projection_plane, neurite_type)
    return np.nan if len(projected_points) == 0 else morphmath.aspect_ratio(projected_points)


@feature(shape=())
def circularity(morph, neurite_type=NeuriteType.all, projection_plane="xy"):
    """Calculates the circularity of the morphology points along the plane.

    The circularity is defined as the 4 * pi * area of the convex hull over its
    perimeter.

    Args:
        morph: Morphology object.
        neurite_type: The neurite type to use. By default all neurite types are used.
        projection_plane: Projection plane to use for the calculation. One of
            ('xy', 'xz', 'yz').

    Returns:
        The circularity of the morphology points.
    """
    projected_points = _unique_projected_points(morph, projection_plane, neurite_type)
    return np.nan if len(projected_points) == 0 else morphmath.circularity(projected_points)


@feature(shape=())
def shape_factor(morph, neurite_type=NeuriteType.all, projection_plane="xy"):
    """Calculates the shape factor of the morphology points along the plane.

    The shape factor is defined as the ratio of the convex hull area over max squared
    pairwise distance of the morphology points.

    Args:
        morph: Morphology object.
        neurite_type: The neurite type to use. By default all neurite types are used.
        projection_plane: Projection plane to use for the calculation. One of
            ('xy', 'xz', 'yz').

    Returns:
        The shape factor of the morphology points.
    """
    projected_points = _unique_projected_points(morph, projection_plane, neurite_type)
    return np.nan if len(projected_points) == 0 else morphmath.shape_factor(projected_points)


@feature(shape=())
def length_fraction_above_soma(morph, neurite_type=NeuriteType.all, up="Y"):
    """Returns the length fraction of the segments that have their midpoints higher than the soma.

    Args:
        morph: Morphology object.
        neurite_type: The neurite type to use. By default all neurite types are used.
        up: The axis along which the computation is performed. One of ('X', 'Y', 'Z').

    Returns:
        The fraction of neurite length that lies on the right of the soma along the given axis.
    """
    _assert_soma_center(morph)
    axis = up.upper()

    if axis not in {"X", "Y", "Z"}:
        raise NeuroMError(f"Unknown axis {axis}. Please choose 'X', 'Y', or 'Z'.")

    col = getattr(COLS, axis)

    segments = _get_segments(morph, neurite_type)

    if not segments:
        return np.nan

    # (Segment 1, Segment 2) x (X, Y, Z, R) X N
    segments = np.dstack(segments)

    # shape N x 3
    seg_begs = segments[0, COLS.XYZ, :].T
    seg_ends = segments[1, COLS.XYZ, :].T

    lengths = np.linalg.norm(seg_begs - seg_ends, axis=1)

    midpoints = 0.5 * (seg_begs + seg_ends)
    selection = midpoints[:, col] > morph.soma.center[col]

    return lengths[selection].sum() / lengths.sum()
